;; -*- Mode: Lisp; Syntax: ANSI-Common-Lisp -*-

(in-package :iolib-posix)

;; sizes of standard types
(defconstant size-of-pointer (foreign-type-size :pointer))
(defconstant size-of-char (foreign-type-size :char))
(defconstant size-of-short (foreign-type-size :short))
(defconstant size-of-int (foreign-type-size :int))
(defconstant size-of-long (foreign-type-size :long))
(defconstant size-of-long-long (foreign-type-size :long-long))
(mapc #'export '(size-of-pointer size-of-char size-of-short
                 size-of-int size-of-long size-of-long-long))

(export 'foreign-pointer)
(deftype foreign-pointer ()
  '#.(type-of (null-pointer)))


;;;;;;;;;;;;;;;;
;;;          ;;;
;;; stdlib.h ;;;
;;;          ;;;
;;;;;;;;;;;;;;;;

(defsyscall ("mkstemp" %mkstemp) :int :noexport
  (template :pointer))

(define-entry-point mkstemp (template)
  (let ((new-template
         (concatenate 'string template "XXXXXX"))
        retval)
    (with-foreign-string (cstr new-template)
      (setf retval (%mkstemp cstr))
      (values retval (foreign-string-to-lisp cstr)))))

(defsyscall ("mkdtemp" %mkdtemp) :string :noexport
  (template :string))

(define-entry-point mkdtemp (template)
  (let ((new-template
         (concatenate 'string template "XXXXXX")))
    (%mkdtemp new-template)))


;;;;;;;;;;;;;;;;
;;;          ;;;
;;; string.h ;;;
;;;          ;;;
;;;;;;;;;;;;;;;;

(defsyscall "memset" :pointer
  (buffer :pointer)
  (value  :int)
  (length :size))

(define-entry-point bzero (buffer length)
  (memset buffer 0 length))

(defsyscall "memcpy" :pointer
  (dest   :pointer)
  (src    :pointer)
  (length :size))

(defsyscall "memmove" :pointer
  (dest   :pointer)
  (src    :pointer)
  (length :size))


;;;;;;;;;;;;;;;;
;;;          ;;;
;;; unistd.h ;;;
;;;          ;;;
;;;;;;;;;;;;;;;;

;;
;; I/O
;;
(defsyscall "read" :ssize
  "Read at most count bytes from fd into the foreign area buf."
  (fd    :int)
  (buf   :pointer)
  (count :size))

(defsyscall "write" :ssize
  "Write at most count bytes to fd from the foreign area buf."
  (fd    :int)
  (buf   :pointer)
  (count :size))

(defsyscall ("pread" pread :wrap) :ssize
  (fd     :int)
  (buf    :pointer)
  (count  :size)
  (offset :off))

(defsyscall ("pwrite" pwrite :wrap) :ssize
  (fd     :int)
  (buf    :pointer)
  (count  :size)
  (offset :off))

;;
;; directories
;;
(defsyscall "chdir" :int
  "Changes the current working directory"
  (path :string))

(defsyscall "fchdir" :int
  (fildes :int))

(defsyscall ("getcwd" %getcwd) :string :noexport
  "Returns the current working directory as a string."
  (buf :pointer)
  (size :size))

(define-entry-point getcwd ()
  "Returns the current working directory as a string."
  (with-foreign-pointer (buf path-max size)
    (%getcwd buf size)))

(defsyscall "link" :int
  (path1 :string)
  (path2 :string))

(defsyscall "rmdir" :int
  (path :string))

;;
;; files
;;
(defsyscall ("open" %open-without-mode) :int :noexport
  (pathname :string)
  (flags    :int))

(defsyscall ("open" %open-with-mode) :int :noexport
  (pathname :string)
  (flags    :int)
  (mode     :mode))

(define-entry-point open (pathname flags &optional (mode nil modep))
  (cond ((not modep)      (%open-without-mode pathname flags))
        ((integerp modep) (%open-with-mode    pathname flags mode))
        (t                (error "Wrong mode: ~S" mode))))

(defsyscall "creat" :int
  (pathname :string)
  (mode     :mode))

(defsyscall ("pipe" %pipe) :int :noexport
  "Create pipe."
  (filedes :pointer))

(define-entry-point pipe ()
  (with-foreign-object (filedes :int 2)
    (%pipe filedes)
    (values (mem-aref filedes :int 0)
            (mem-aref filedes :int 1))))

(defsyscall "rename" :int
  "Rename a file."
  (old :string)
  (new :string))

(defsyscall "close" :int
  "Close an open file descriptor."
  (fd :int))

(defsyscall ("readlink" %readlink) :ssize :noexport
  (path    :string)
  (buf     :pointer)
  (bufsize :size))

(define-entry-point readlink (path)
  (with-foreign-pointer-as-string (buf 4096 bufsize)
    (%readlink path buf bufsize)
    buf))

(defsyscall "symlink" :int
  "Creates a symbolic link"
  (name1 :string)
  (name2 :string))

(defsyscall "unlink" :int
  (path :string))

(defsyscall "access" :int
  (path  :string)
  (amode :int))

(defsyscall "dup" :int
  (fildes :int))

(defsyscall "dup2" :int
  (fildes1 :int)
  (fildes2 :int))

(defsyscall ("truncate" truncate :wrap) :int
  (path   :string)
  (length :off))

(defsyscall ("ftruncate" ftruncate :wrap) :int
  (fildes :int)
  (length :off))

(defsyscall ("lseek" lseek :wrap) :off
  (fildes :int)
  (offset :off)
  (whence :int))

(defsyscall ("chown" %chown) :int :noexport
  "Change ownership of a file."
  (path  :string)
  (owner :uid)
  (group :uid))

(defsyscall ("lchown" %lchown) :int :noexport
  "Change ownership of a file or symlink."
  (fd    :int)
  (owner :uid)
  (group :uid))

(defsyscall ("fchown" %fchown) :int :noexport
  "Change ownership of an open file."
  (path  :string)
  (owner :uid)
  (group :uid))

(define-entry-point change-owner (path-or-fd &optional owner group (follow-symlink t))
  "Change ownership of a file."
  (unless (or owner group)
    (error "At least one of OWNER and GROUP must be specified."))
  (let ((c-owner (or owner -1))
        (c-group (or group -1)))
      (cond ((integerp path-or-fd) (%fchown path-or-fd c-owner c-group))
            (follow-symlink        (%chown  path-or-fd c-owner c-group))
            (t                     (%lchown path-or-fd c-owner c-group)))))

(defsyscall "sync" :void
  "Schedule all file system buffers to be written to disk.")

(defsyscall "fsync" :int
  (fildes :int))

#+linux
(defsyscall "fdatasync" :int
  (fd :int))

(defsyscall "lockf" :int
  "Apply, test or remove a POSIX lock on an open file."
  (fd  :int)
  (cmd :int)
  (len :off))

;;
;; processes
;;

(defforeign ("nice" %nice) :int :noexport
  "Change process priority."
  (increment :int))

(define-entry-point nice (increment)
  "Change process priority."
  (set-errno 0)
  (let ((r (%nice increment)))
    (if (and (= r -1) (/= (get-errno) 0))
        (signal-unix-error)
        r)))

(defsyscall "fork" :pid
  "Create a child process.")

(defsyscall "getegid" :gid
  "Get effective group id of the current process.")

(defsyscall "geteuid" :uid
  "Get effective user id of the current process.")

(defsyscall "getgid" :gid
  "Get real group id of the current process.")

(defsyscall "getpgid" :pid
  "Get process group id of a process."
  (pid :pid))

(defsyscall "getpgrp" :pid
  "Get process group id of the current process.")

(defsyscall "getpid" :pid
  "Returns the process id of the current process")

(defsyscall "getppid" :pid
  "Returns the process id of the current process's parent")

(defsyscall "getuid" :uid
  "Get real user id of the current process.")

(defsyscall "setegid" :int
  "Set effective group id of the current process."
  (gid :gid))

(defsyscall "seteuid" :int
  "Set effective user id of the current process."
  (uid :uid))

(defsyscall "setgid" :int
  "Get real group id of the current process."
  (gid :gid))

(defsyscall "setpgid" :int
  "Set process group id of a process."
  (pid  :pid)
  (pgid :pid))

(defsyscall "setpgrp" :pid
  "Set process group id of the current process.")

(defsyscall "setregid" :int
  "Set real and effective group id of the current process."
  (rgid :gid)
  (egid :gid))

(defsyscall "setreuid" :int
  "Set real and effective user id of the current process."
  (ruid :uid)
  (euid :uid))

(defsyscall "setsid" :pid
  "Create session and set process group id of the current process.")

(defsyscall "setuid" :int
  "Set real user id of the current process."
  (uid :uid))

;;
;; environment
;;
(defsyscall "getenv" :string
  "Returns the value of an environment variable"
  (name :string))

(defsyscall "setenv" :int
  "Changes the value of an environment variable"
  (name  :string)
  (value :string))

(defsyscall "unsetenv" :int
  "Removes the binding of an environment variable"
  (name :string))

;;
;; time
;;
(defsyscall "sleep" :unsigned-int
  (seconds :unsigned-int))

(defsyscall "usleep" :int
  (useconds :useconds))

;;
;; misc
;;

(defsyscall ("gethostname" %gethostname) :int :noexport
  (name    :pointer)
  (namelen :size))

(define-entry-point get-host-name ()
  (with-foreign-pointer-as-string (cstr 256 size)
    (%gethostname cstr size)
    (foreign-string-to-lisp cstr)))

(defsyscall ("getdomainname" %getdomainname) :int :noexport
  (name    :pointer)
  (namelen :size))

(define-entry-point get-domain-name ()
  (with-foreign-pointer-as-string (cstr 256 size)
    (%getdomainname cstr size)
    (foreign-string-to-lisp cstr)))

(defsyscall ("sysconf" %sysconf) :long :noexport
  (name :int))

(define-entry-point sysconf (name)
  "Get value of configurable system variables."
  (set-errno 0)
  (let ((r (%sysconf name)))
    (if (and (= r -1) (/= (get-errno) 0))
        (signal-unix-error)
        r)))

(define-entry-point getpagesize ()
  "Get memory page size."
  (sysconf SC-PAGESIZE))


;;;;;;;;;;;;;;;
;;;         ;;;
;;; fcntl.h ;;;
;;;         ;;;
;;;;;;;;;;;;;;;

(defsyscall ("fcntl" %fcntl-without-arg) :int :noexport
  (fd  :int)
  (cmd :int))

(defsyscall ("fcntl" %fcntl-with-int-arg) :int :noexport
  (fd  :int)
  (cmd :int)
  (arg :int))

(defsyscall ("fcntl" %fcntl-with-pointer-arg) :int :noexport
  (fd  :int)
  (cmd :int)
  (arg :pointer))

(define-entry-point fcntl (fd cmd &optional (arg nil argp))
  (cond ((not argp)     (%fcntl-without-arg fd cmd))
        ((integerp arg) (%fcntl-with-int-arg fd cmd arg))
        ((pointerp arg) (%fcntl-with-pointer-arg fd cmd arg))
        (t              (error "Wrong argument to fcntl: ~S" arg))))


;;;;;;;;;;;;;;;
;;;         ;;;
;;; ioctl.h ;;;
;;;         ;;;
;;;;;;;;;;;;;;;

(defsyscall ("ioctl" %ioctl-without-arg) :int :noexport
 "Control device."
 (fd      :int)
 (request :int))

(defsyscall ("ioctl" %ioctl-with-arg) :int :noexport
 "Control device."
 (fd      :int)
 (request :int)
 (arg     :pointer))

(define-entry-point ioctl (fd request &optional (arg nil argp))
  (cond ((not argp)     (%ioctl-without-arg fd request))
        ((pointerp arg) (%ioctl-with-arg fd request arg))
        (t              (error "Wrong argument to fcntl: ~S" arg))))


;;;;;;;;;;;;;;;;;;
;;;            ;;;
;;; sys/mman.h ;;;
;;;            ;;;
;;;;;;;;;;;;;;;;;;

(defsyscall "mlock" :int
  "Lock a range of process address space."
  (addr :pointer)
  (len  :size))

(defsyscall "munlock" :int
  "Unlock a range of process address space."
  (addr :pointer)
  (len  :size))

(defsyscall "mlockall" :int
  "Lock the address space of a process."
  (flags :int))

(defsyscall "munlockall" :int)

(defsyscall ("mmap" mmap :wrap) :pointer
  "Map pages of memory."
  (addr    :pointer)
  (len     :size)
  (prot    :int)
  (flags   :int)
  (filedes :int)
  (off     :off))

(defsyscall "munmap" :int
  "Unmap pages of memory."
  (addr :pointer)
  (len  :size))

(defsyscall "msync" :int
  "Synchronize memory with physical storage."
  (addr  :pointer)
  (len   :size)
  (flags :int))

(defsyscall "mprotect" :int
  "Set protection of memory mapping."
  (addr  :pointer)
  (len   :size)
  (flags :int))


;;;;;;;;;;;;;;;;;;
;;;            ;;;
;;; sys/time.h ;;;
;;;            ;;;
;;;;;;;;;;;;;;;;;;

(defsyscall ("gettimeofday" %gettimeofday) :int :noexport
  (tp  :pointer)
  (tzp :pointer))

(define-entry-point gettimeofday ()
  "Return the time in seconds and microseconds."
  (with-foreign-object (tv 'timeval)
    (with-foreign-slots ((sec usec) tv timeval)
      (%gettimeofday tv (null-pointer))
      (values sec usec))))

(defsyscall ("clock_getres" %clock-getres) :int :noexport
  "Returns the resolution of the clock CLOCKID."
  (clockid :clockid)
  (res     :pointer))

(define-entry-point clock-get-resolution (clock-id)
  (with-foreign-object (ts 'timespec)
    (with-foreign-slots ((sec nsec) ts timespec)
      (%clock-getres clock-id ts)
      (values sec nsec))))

(defsyscall ("clock_gettime" %clock-gettime) :int :noexport
  "Returns the time of the clock CLOCKID."
  (clockid :clockid)
  (tp      :pointer))

(define-entry-point clock-get-time (clock-id)
  (with-foreign-object (ts 'timespec)
    (with-foreign-slots ((sec nsec) ts timespec)
      (%clock-gettime clock-id ts)
      (values sec nsec))))

(defsyscall ("clock_settime" %clock-settime) :int :noexport
  "Sets the time of the clock CLOCKID."
  (clockid :clockid)
  (tp      :pointer))

(define-entry-point clock-set-time (clock-id)
  (with-foreign-object (ts 'timespec)
    (with-foreign-slots ((sec nsec) ts timespec)
      (%clock-settime clock-id ts)
      (values sec nsec))))


;;;;;;;;;;;;;;;;;;
;;;            ;;;
;;; sys/stat.h ;;;
;;;            ;;;
;;;;;;;;;;;;;;;;;;

(defsyscall ("stat" %stat :wrap) :int :noexport
  "Get information about a file."
  (path :string)
  (buf  :pointer))

(defsyscall ("lstat" %lstat :wrap) :int :noexport
  "Get information about a file or symlink."
  (path :string)
  (buf  :pointer))

(defsyscall ("fstat" %fstat :wrap) :int :noexport
  "Get information about a file descriptor."
  (fd  :int)
  (buf :pointer))

(define-entry-point stat (path-or-fd &optional (follow-symlink t))
  "Get information about a filesystem object."
  (with-foreign-object (buf 'stat)
    (cond ((integerp path-or-fd) (%fstat path-or-fd buf))
          (follow-symlink        (%stat  path-or-fd buf))
          (t                     (%lstat path-or-fd buf)))
    (with-foreign-slots ((dev ino mode nlink uid gid
                         rdev size blksize blocks
                         atime mtime ctime)
                         buf stat)
      (values dev ino mode nlink uid gid
              rdev size blksize blocks
              atime mtime ctime))))

(defsyscall "umask" :mode
  "Sets the umask and returns the old one"
  (new-mode :mode))

(defsyscall "mkdir" :int
  "Create a directory."
  (path :string)
  (mode :mode))

(defsyscall ("mknod" mknod :wrap) :int
  "Create a special or an ordinary file."
  (path :string)
  (mode :mode)
  (dev  :dev))

(defsyscall "mkfifo" :int
  "Create a FIFO(named pipe)."
  (path :string)
  (mode :mode))


;;;;;;;;;;;;;;;;;;;;
;;;              ;;;
;;; sys/select.h ;;;
;;;              ;;;
;;;;;;;;;;;;;;;;;;;;

(defsyscall "select" :int
  "Scan for I/O activity on multiple file descriptors."
  (nfds      :int)
  (readfds   :pointer)
  (writefds  :pointer)
  (exceptfds :pointer)
  (timeout   :pointer))

(declaim (inline fd-zero fd-clr fd-set fd-isset))

(define-entry-point fd-zero (fd-set)
  (declare (optimize (speed 3) (safety 0) (space 0) (debug 0))
           (type foreign-pointer fd-set))
  (bzero fd-set size-of-fd-set)
  (values fd-set))

(define-entry-point copy-fd-set (from to)
  (declare (optimize (speed 3) (safety 0) (space 0) (debug 0))
           (type foreign-pointer from to))
  (memcpy to from size-of-fd-set)
  (values to))

(export 'select-file-descriptor)
(deftype select-file-descriptor ()
  `(mod #.fd-setsize))

(define-entry-point fd-isset (fd fd-set)
  (declare (optimize (speed 3) (safety 0) (space 0) (debug 0))
           (type select-file-descriptor fd)
           (type foreign-pointer fd-set))
  (multiple-value-bind (byte-off bit-off) (floor fd 8)
    (let ((oldval (mem-aref fd-set :uint8 byte-off)))
      (logbitp bit-off oldval))))

(define-entry-point fd-clr (fd fd-set)
  (declare (optimize (speed 3) (safety 0) (space 0) (debug 0))
           (type select-file-descriptor fd)
           (type foreign-pointer fd-set))
  (multiple-value-bind (byte-off bit-off) (floor fd 8)
    (let ((oldval (mem-aref fd-set :uint8 byte-off)))
      (setf (mem-aref fd-set :uint8 byte-off)
            (logandc2 oldval (ash 1 bit-off)))))
  (values fd-set))

(define-entry-point fd-set (fd fd-set)
  (declare (optimize (speed 3) (safety 0) (space 0) (debug 0))
           (type select-file-descriptor fd)
           (type foreign-pointer fd-set))
  (multiple-value-bind (byte-off bit-off) (floor fd 8)
    (let ((oldval (mem-aref fd-set :uint8 byte-off)))
      (setf (mem-aref fd-set :uint8 byte-off)
            (logior oldval (ash 1 bit-off)))))
  (values fd-set))


;;;;;;;;;;;;;;;;
;;;          ;;;
;;; syslog.h ;;;
;;;          ;;;
;;;;;;;;;;;;;;;;

(defvar *syslog-identity* nil)

(defsyscall ("openlog" %openlog) :void :noexport
  "Open system log."
  (ident    :string)
  (option   :int)
  (facility :int))

;; not my BUG: on Linux openlog() keeps a pointer to the "ident"
;; string which syslog() then uses when logging, therefore
;; we need to keep the foreign string around
(define-entry-point openlog (name facility &optional (option 0))
  (when *syslog-identity*
    (foreign-string-free *syslog-identity*)
    (setf *syslog-identity* nil))
  (setf *syslog-identity* (foreign-string-alloc name))
  (%openlog *syslog-identity* option
            (foreign-enum-value 'syslog-facilities facility)))

(defsyscall ("syslog" %syslog) :void :noexport
  (priority :int)
  (format   :string)
  (message  :string))

(define-entry-point syslog (priority message)
  (%syslog (foreign-enum-value 'syslog-priorities priority)
           "%s" message))

(defsyscall ("closelog" %closelog) :void :noexport
  "Close system log.")

(define-entry-point closelog ()
  (%closelog)
  (when *syslog-identity*
    (foreign-string-free *syslog-identity*)
    (setf *syslog-identity* nil)))

(defsyscall ("setlogmask" set-log-mask) :int
  (mask :int))

(defsyscall ("log_mask" log-mask :wrap) :int
  (priority :int))

(defsyscall ("log_upto" log-upto :wrap) :int
  (priority :int))


;;;;;;;;;;;;;;;;;;;;;;
;;;                ;;;
;;; sys/resource.h ;;;
;;;                ;;;
;;;;;;;;;;;;;;;;;;;;;;

(defsyscall ("getrlimit" %getrlimit :wrap) :int :noexport
  (resource :int)
  (rlimit   :pointer))

(define-entry-point get-resource-limit (resource)
  (with-foreign-object (rl 'rlimit)
    (with-foreign-slots ((cur max) rl rlimit)
      (%getrlimit resource rl)
      (values cur max))))

(defsyscall ("setrlimit" %setrlimit :wrap) :int :noexport
  (resource :int)
  (rlimit   :pointer))

(define-entry-point set-resource-limit (resource soft-limit hard-limit)
  (with-foreign-object (rl 'rlimit)
    (with-foreign-slots ((cur max) rl rlimit)
      (setf cur soft-limit
            max hard-limit)
      (%setrlimit resource rl))))

(defsyscall ("getrusage" %getrusage) :int :noexport
  (who   :int)
  (usage :pointer))

(define-entry-point get-resource-usage (who)
  (with-foreign-object (ru 'rusage)
    (%getrusage who ru)
    (with-foreign-slots ((maxrss ixrss
                          idrss isrss minflt majflt
                          nswap inblock oublock msgsnd
                          msgrcv nsignals nvcsw nivcsw)
                         ru rusage)
      (values (foreign-slot-value (foreign-slot-pointer ru 'rusage 'utime)
                                  'timeval 'sec)
              (foreign-slot-value (foreign-slot-pointer ru 'rusage 'utime)
                                  'timeval 'usec)
              (foreign-slot-value (foreign-slot-pointer ru 'rusage 'stime)
                                  'timeval 'sec)
              (foreign-slot-value (foreign-slot-pointer ru 'rusage 'stime)
                                  'timeval 'usec)
              maxrss ixrss idrss isrss minflt majflt
              nswap inblock oublock msgsnd
              msgrcv nsignals nvcsw nivcsw))))

(defsyscall "getpriority" :int
  (which :int)
  (who   :int))

(defsyscall "setpriority" :int
  (which :int)
  (who   :int)
  (value :int))


;;;;;;;;;;;;;;;;;;;;;
;;;               ;;;
;;; sys/utsname.h ;;;
;;;               ;;;
;;;;;;;;;;;;;;;;;;;;;

(defsyscall ("uname" %uname) :int :noexport
  "Get name and information about current kernel."
  (buf :pointer))

(define-entry-point uname ()
  (with-foreign-object (buf 'utsname)
    (%uname buf)
    (macrolet ((utsname-slot (name)
                 `(foreign-string-to-lisp
                   (foreign-slot-pointer buf 'utsname ',name))))
      (values (utsname-slot sysname)
              (utsname-slot nodename)
              (utsname-slot release)
              (utsname-slot version)
              (utsname-slot machine)))))


;;;;;;;;;;;;;;;;;;;;;
;;;               ;;;
;;; sys/statvfs.h ;;;
;;;               ;;;
;;;;;;;;;;;;;;;;;;;;;


(defsyscall ("statvfs" %statvfs :wrap) :int :noexport
  "Get file system statistics."
  (path :string)
  (buf  :pointer))

(defsyscall ("fstatvfs" %fstatvfs :wrap) :int :noexport
  "Get file system statistics."
  (fd  :int)
  (buf :pointer))

(define-entry-point statvfs (path-or-fd)
  "Get information about a filesystem."
  (with-foreign-object (buf 'statvfs)
    (cond ((integerp path-or-fd) (%fstatvfs path-or-fd buf))
          ((stringp path-or-fd)  (%statvfs  path-or-fd buf))
          (t                     (error "Wrong argument to statvfs: ~S" path-or-fd)))
    (with-foreign-slots ((bsize frsize blocks bfree
                          bavail files ffree favail
                          fsig flag namemax)
                         buf statvfs)
      (values bsize frsize blocks bfree
              bavail files ffree favail
              fsig flag namemax))))


;;;;;;;;;;;;;;;;;;
;;;            ;;;
;;; sys/poll.h ;;;
;;;            ;;;
;;;;;;;;;;;;;;;;;;

(defsyscall "poll" :int
  "Scan for I/O activity on multiple file descriptors."
  (fds     :pointer)
  (nfds    :nfds)
  (timeout :int))


;;;;;;;;;;;;;;;;;;;
;;;             ;;;
;;; sys/epoll.h ;;;
;;;             ;;;
;;;;;;;;;;;;;;;;;;;

#+linux
(defsyscall "epoll_create" :int
  "Open an epoll file descriptor."
  (size :int))

#+linux
(defsyscall "epoll_ctl" :int
  "Control interface for an epoll descriptor."
  (epfd  :int)
  (op    :int)
  (fd    :int)
  (event :pointer))

#+linux
(defsyscall "epoll_wait" :int
  "Wait for an I/O event on an epoll file descriptor."
  (epfd      :int)
  (events    :pointer)
  (maxevents :int)
  (timeout   :int))


;;;;;;;;;;;;;;;;;;;
;;;             ;;;
;;; sys/event.h ;;;
;;;             ;;;
;;;;;;;;;;;;;;;;;;;

#+freebsd
(defsyscall "kqueue" :int
  "Open a kernel event queue.")

#+freebsd
(defsyscall "kevent" :int
  "Control interface for a kernel event queue."
  (kq         :int)
  (changelist :pointer) ; const struct kevent *
  (nchanges   :int)
  (eventlist  :pointer) ; struct kevent *
  (nevents    :int)
  (timeout    :pointer)) ; const struct timespec *

#+freebsd
(define-entry-point ev-set (_kev _ident _filter _flags _fflags _data _udata)
  (with-foreign-slots ((ident filter flags fflags data udata) _kev kevent)
    (setf ident  _ident  filter _filter flags _flags
          fflags _fflags data   _data   udata _udata)))


;;;;;;;;;;;;;
;;;       ;;;
;;; pwd.h ;;;
;;;       ;;;
;;;;;;;;;;;;;

(defsyscall ("getpwuid_r" %getpwuid-r) :int :noexport
  "Gets the password-entry of a user, by user id."
  (uid     :uid)
  (pwd     :pointer)
  (buffer  :pointer)
  (bufsize :size)
  (result  :pointer))

(defsyscall ("getpwnam_r" %getpwnam-r) :int :noexport
  "Gets the password-entry of a user, by username."
  (name    :string)
  (pwd     :pointer)
  (buffer  :pointer)
  (bufsize :size)
  (result  :pointer))

(define-entry-point get-passwd-entry (name-or-id)
  "Gets the password-entry of a user."
  (with-foreign-objects ((pw 'passwd)
                         (pwp :pointer))
    (with-foreign-pointer (buf 4096 bufsize)
      (with-foreign-slots ((name passwd uid gid
                            gecos dir shell)
                           pw passwd)
        (let ((res (etypecase name-or-id
                     (integer (%getpwuid-r name-or-id pw buf bufsize pwp))
                     (string  (%getpwnam-r name-or-id pw buf bufsize pwp)))))
          (if (and (zerop res)
                   (null-pointer-p (mem-ref pwp :pointer)))
              (values)
              (values name passwd uid gid gecos dir shell)))))))


;;;;;;;;;;;;;
;;;       ;;;
;;; grp.h ;;;
;;;       ;;;
;;;;;;;;;;;;;

(defsyscall ("getgrgid_r" %getgrgid-r) :int :noexport
  "Gets a group-entry, by group id."
  (uid     :uid)
  (grp     :pointer)
  (buffer  :pointer)
  (bufsize :size)
  (result  :pointer))

(defsyscall ("getgrnam_r" %getgrnam-r) :int :noexport
  "Gets a group-entry, by group name."
  (name    :string)
  (grp     :pointer)
  (buffer  :pointer)
  (bufsize :size)
  (result  :pointer))

(define-entry-point get-group-entry (name-or-id)
  "Gets a group entry."
  (with-foreign-objects ((gr 'group)
                         (grp :pointer))
    (with-foreign-pointer (buf 4096 bufsize)
      (with-foreign-slots ((name passwd gid)
                           gr group)
        (let ((res (etypecase name-or-id
                     (integer (%getgrgid-r name-or-id gr buf bufsize grp))
                     (string  (%getgrnam-r name-or-id gr buf bufsize grp)))))
          (if (and (zerop res)
                   (null-pointer-p (mem-ref grp :pointer)))
              (values)
              (values name passwd gid)))))))


;;;;;;;;;;;;;;;;
;;;          ;;;
;;; dirent.h ;;;
;;;          ;;;
;;;;;;;;;;;;;;;;

(defsyscall "opendir" :pointer
  "Opens a directory for listing of its contents"
  (filename :string))

(defsyscall "closedir" :int
  "Closes a directory when done listing its contents"
  (dir :pointer))

(defsyscall ("readdir_r" %readdir-r :wrap) :int :noexport
  "Reads an item from the listing of a directory(reentrant)"
  (dir    :pointer)
  (entry  :pointer)
  (result :pointer))

(define-entry-point readdir (dir)
  (with-foreign-objects ((entry 'dirent)
                         (result :pointer))
    (%readdir-r dir entry result)
    (if (null-pointer-p (mem-ref result :pointer))
        nil
        (with-foreign-slots ((name type fileno) entry dirent)
          (values (foreign-string-to-lisp name)
                  type fileno)))))

(defsyscall "rewinddir" :void
  "Rewinds a directory"
  (dir :pointer))

(defsyscall "seekdir" :void
  "Seeks a directory"
  (dir :pointer)
  (pos :long))

;; according to POSIX docs "no errors are defined" for telldir but
;; Linux manpages specify a possible EBADF
(defsyscall "telldir" :off
  "Returns the current location in a directory"
  (dir :pointer))

#-freebsd
(defsyscall "dirfd" :int
  "Returns the file descriptor of a directory"
  (dir :pointer))


;;;;;;;;;;;;;;;;;
;;;           ;;;
;;; sys/uio.h ;;;
;;;           ;;;
;;;;;;;;;;;;;;;;;

(defsyscall "readv" :ssize
  (fd    :int)
  (iov   :pointer)
  (count :size))

(defsyscall "writev" :ssize
  (fd    :int)
  (iov   :pointer)
  (count :size))


;;;;;;;;;;;;;;;;;;;;
;;;              ;;;
;;; sys/socket.h ;;;
;;;              ;;;
;;;;;;;;;;;;;;;;;;;;

(defsyscall "accept" :int
  "Accept an incoming connection, returning the file descriptor."
  (socket  :int)
  (address :pointer) ; sockaddr-foo
  (addrlen :pointer))

(defsyscall "bind" :int
  "Bind a socket to a particular local address."
  (fd      :int)
  (address :pointer)
  (addrlen :socklen))

(defsyscall "connect" :int
  "Create an outgoing connection on a given socket."
  (socket  :int)
  (address :pointer) ; sockaddr-foo
  (addrlen :socklen))

(defsyscall "getpeername" :int
  (socket  :int)
  (address :pointer)
  (addrlen :pointer))

(defsyscall "getsockname" :int
  (socket  :int)
  (address :pointer)
  (addrlen :pointer))

(defsyscall "getsockopt" :int
  "Retrieve socket configuration."
  (fd      :int)
  (level   :int)
  (optname :int)
  (optval  :pointer)
  (optlen  :pointer))

(defsyscall "listen" :int
  "Mark a bound socket as listening for incoming connections."
  (socket  :int)
  (backlog :int))

(defsyscall "recv" :ssize
  (socket :int)
  (buffer :pointer)
  (length :size)
  (flags  :int))

(defsyscall "recvfrom" :ssize
  (socket  :int)
  (buffer  :pointer)
  (length  :size)
  (flags   :int)
  (address :pointer)
  (addrlen :pointer))

(defsyscall "recvmsg" :ssize
  (socket  :int)
  (message :pointer)
  (flags   :int))

(defsyscall "send" :ssize
  (socket :int)
  (buffer :pointer)
  (length :size)
  (flags  :int))

(defsyscall "sendmsg" :ssize
  (socket  :int)
  (message :pointer)
  (flags   :int))

(defsyscall "sendto" :ssize
  (socket   :int)
  (buffer   :pointer)
  (length   :size)
  (flags    :int)
  (destaddr :pointer)
  (destlen  :socklen))

(defsyscall "setsockopt" :int
  "Configure a socket."
  (fd      :int)
  (level   :int)
  (optname :int)
  (optval  :pointer)
  (optlen  :socklen))

(defsyscall "shutdown" :int
  (socket :int)
  (how    :int))

(defsyscall "socket" :int
  "Create a BSD socket."
  (domain   :int)  ; af-*
  (type     :int)  ; sock-*
  (protocol :int))

(defsyscall "sockatmark" :int
  (socket :int))

(defsyscall ("socketpair" %socketpair) :int :noexport
  (domain   :int)  ; af-*
  (type     :int)  ; sock-*
  (protocol :int)
  (filedes  :pointer))

(define-entry-point socketpair (domain type protocol)
  (with-foreign-object (filedes :int 2)
    (%socketpair domain type protocol filedes)
    (values (mem-aref filedes :int 0)
            (mem-aref filedes :int 1))))


;;;;;;;;;;;;;;;;;;;;
;;;              ;;;
;;; netinet/un.h ;;;
;;;              ;;;
;;;;;;;;;;;;;;;;;;;;

(export 'unix-path-max)
(defconstant unix-path-max (- size-of-sockaddr-un
                              (foreign-slot-offset 'sockaddr-un 'path)))

;;;;;;;;;;;;;;;
;;;         ;;;
;;; netdb.h ;;;
;;;         ;;;
;;;;;;;;;;;;;;;

(defsyscall "gai_strerror" :string
  (errcode :int))

(defsyscall "freeaddrinfo" :void
  (ai :pointer))

(define-condition resolv-error (system-error) ())

(defmethod print-object ((resolv-error resolv-error) stream)
  (print-unreadable-object (resolv-error stream :type t :identity nil)
    (let ((code (system-error-code resolv-error))
          (identifier (system-error-identifier resolv-error)))
      (format stream "~s ~s ~s"
              (or code "[No code]") identifier
              (or (ignore-errors (gai-strerror code))
		  "[Can't get error string.]")))))

(defun resolv-error (retval)
  (let ((identifier (foreign-enum-keyword 'addrinfo-errors retval)))
    (case identifier
      (:eai-system (signal-unix-error))
      (otherwise (error 'resolv-error
                        :code retval
                        :identifier identifier)))))

(defforeign-with-helpers "getaddrinfo" :int
  (:error-predicate (lambda (x) (not (zerop x)))
   :filter identity
   :error-generator resolv-error)
  (node    :string)
  (service :string)
  (hints   :pointer)
  (result  :pointer))

(defforeign-with-helpers "getnameinfo" :int
  (:error-predicate (lambda (x) (not (zerop x)))
   :filter identity
   :error-generator resolv-error)
  (sa         :pointer)
  (salen      :socklen)
  (node       :pointer)
  (nodelen    :socklen)
  (service    :pointer)
  (servicelen :socklen)
  (flags      :int))

(defsyscall "getprotobyname" :pointer
  (name :string))

(defsyscall "getprotobynumber" :pointer
  (proto :int))


;;;;;;;;;;;;;;;;;;;
;;;             ;;;
;;; arpa/inet.h ;;;
;;;             ;;;
;;;;;;;;;;;;;;;;;;;

(defsyscall "inet_ntop" :string
  (family :int)
  (src    :pointer)
  (dest   :pointer)
  (size   :socklen))

(defforeign-with-helpers "inet_pton" :int
  (:error-predicate (lambda (x) (not (plusp x)))
   :filter identity
   :error-generator signal-unix-error)
  (family :int)
  (src    :string)
  (dest   :pointer))


;;;;;;;;;;;;;;;;
;;;          ;;;
;;; net/if.h ;;;
;;;          ;;;
;;;;;;;;;;;;;;;;

(defforeign-with-helpers "if_nametoindex" :unsigned-int
  (:error-predicate zerop
   :filter identity
   :error-generator signal-unix-error)
  (ifname :string))

(defsyscall "if_indextoname" :string
  (ifindex :unsigned-int)
  (ifname  :pointer))

(defsyscall "if_nameindex" :pointer
  "Return all network interface names and indexes")

(defsyscall "if_freenameindex" :void
  (ptr :pointer))
